-- Sonic_Minefield_Building_Effect
-- Author: HandyVac
-- DateCreated: 8/23/2015 3:38:11 PM
--------------------------------------------------------------
-- this script implements the effect of the Ultrasonic Minefield building.
-- causes damage each turn to any alien unit within 2 tiles of a city with this building.

local BuildingSonicMinefield_ID = GameInfo.Buildings["BUILDING_ULTRASONIC_MINEFIELD"].ID;
local iDamageAmount = 10;
local ExtraRangePerk_ID = GameInfo.PlayerPerks["PLAYERPERK_ULTRASONIC_MINEFIELD_EXTRA_RANGE"].ID;
local PortableEmittersPerk_ID = GameInfo.PlayerPerks["PLAYERPERK_ULTRASONIC_MINEFIELD_PORTABLE_EMITTERS"].ID;
local PortableEmitterUnit_ID = GameInfo.Units["UNIT_ULTRASONIC_EMITTER"].ID;

function SonicMinefieldEffect (CurrentPlayerID)
	local pPlayer = Players[CurrentPlayerID];

	print("Checking for Sonic Minefield buildings");
	
	-- cycle through all cities this player owns
	for pCity in pPlayer:Cities() do
		--print("checking city: " .. pCity:GetName() .. ", owned by " .. pPlayer:GetName() );

		-- does this city contain a Sonic Minefield building?
		if pCity:IsHasBuilding(BuildingSonicMinefield_ID) then
			print("- found a Sonic Minefield in " .. pCity:GetName() );
			--print("- checking for nearby aliens");

			-- check all plots within a 2 tile radius of the city
			local iRange = 2;
			-- increase range if this player has picked the extra-range building quest option.
			if pPlayer:HasPerk(ExtraRangePerk_ID) then
				iRange = 3;
			end

			for gridX = -iRange, iRange do
				for gridY = -iRange, iRange do
					-- Check whether this plot is within workable range (i.e. a iRange tile *circle* from the city)
					local pPlot = Map.PlotXYWithRangeCheck(pCity:GetX(), pCity:GetY(), gridX, gridY, iRange);
					if pPlot then
						--print(" - checked plot (" .. pPlot:GetX() .. "," .. pPlot:GetY() .. ")");

						-- check if there's a unit here
						if pPlot:IsUnit() then
							local pUnit = pPlot:GetUnit();
											
							-- is it an alien unit?
							if pUnit:GetOwner() > 61 then
								print(" - found an alien " .. pUnit:GetName() .. " in range. Inflicting " .. iDamageAmount .. " damage");
								-- deal damage to the unit
								pUnit:ChangeDamage(iDamageAmount);
							end
						end
					end
				end
			end
		end
	end

end

-- call this at start of every player's turn
GameEvents.PlayerDoTurn.Add(SonicMinefieldEffect);


function DoSonicMinefieldFreeEmitterUnits ()
-- serialevent passes different information than playerdoturn, so cannot rely on the player ID being passed as argument
local CurrentPlayerID = Game.GetActivePlayer()

local pPlayer = Players[CurrentPlayerID];

	-- Has this player picked the portable emitters building quest option?
	if pPlayer:HasPerk(PortableEmittersPerk_ID) then
		-- spawn 3 free emitter units at player's capital 
		local pCity = pPlayer:GetCapitalCity()

		for i = 1, 3 do
			local pUnit = pPlayer:InitUnit( PortableEmitterUnit_ID, pCity:GetX(), pCity:GetY() );
			pUnit:JumpToNearestValidPlot();
		end

		-- remove the playerperk, so this effect only happens once.
		pPlayer:RemovePerk(PortableEmittersPerk_ID);
	end
end



-- call this when the Unit Upgrade window updates graphically. I.e. when the human player is interacting with it.
--Events.SerialEventUnitUpgradeScreenDirty.Add(DoSonicMinefieldFreeEmitterUnits);

-- for the benefit of the human player, call this whenever a popup window is displayed on the screen (e.g. the "Quest Complete" one)
Events.SerialEventGameMessagePopup.Add(DoSonicMinefieldFreeEmitterUnits);


-- and since the AI doesn't use graphical windows, also call this at the start of every player's turn
GameEvents.PlayerDoTurn.Add(DoSonicMinefieldFreeEmitterUnits);